// Copyright 1997-1999 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OFTrie.h"

#import <Foundation/Foundation.h>
#import <OmniBase/OmniBase.h>

#import "OFTrieNode.h"
#import "OFTrieBucket.h"
#import "OFTrieEnumerator.h"

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/OmniFoundation/DataStructures.subproj/OFTrie.m,v 1.7 1999/03/30 21:25:40 kc Exp $")

@implementation OFTrie

// Init and dealloc

- initCaseSensitive:(BOOL)shouldBeCaseSensitive;
{
    if (![super init])
        return nil;

    head = [[OFTrieNode allocWithZone:[self zone]] init];
    caseSensitive = shouldBeCaseSensitive;

    return self;
}

- (void)dealloc;
{
    [head release];
    [super dealloc];
}

//

- (NSEnumerator *)objectEnumerator;
{
    return [[[OFTrieEnumerator alloc] initWithTrie:self] autorelease];
}

- (BOOL)isCaseSensitive;
{
    return caseSensitive;
}

#define SAFE_ALLOCA_SIZE (8 * 8192)

- (void)addBucket:(OFTrieBucket *)bucket forString:(NSString *)aString;
{
    unsigned int length;
    unichar *buffer, *upperBuffer, *ptr;
    OFTrieNode *to, *attachTo = head;
    Class trieNodeClass;
    unsigned int bufferSize;
    BOOL useMalloc;

    length = [aString length];
    bufferSize = (length + 1) * sizeof(unichar);
    useMalloc = bufferSize * 2 >= SAFE_ALLOCA_SIZE;
    if (useMalloc) {
	buffer = (unichar *)NSZoneMalloc(NULL, bufferSize);
	upperBuffer = (unichar *)NSZoneMalloc(NULL, bufferSize);
    } else {
        buffer = (unichar *)alloca(bufferSize);
        upperBuffer = (unichar *)alloca(bufferSize);
    }

    ptr = buffer;

    if (!caseSensitive) {
#warning -addBucket:forString: assumes that -uppercaseString and -lowercaseString return strings of identical length as the original string
        // This isn't actually true for unicode.
        // Also, they assume that string equality is equivalent to having the same unichars in the same sequence, which isn't generally true for unicode.

        aString = [aString uppercaseString];
	[aString getCharacters:upperBuffer];
	upperBuffer[length] = '\0';
	aString = [aString lowercaseString];
    }
    [aString getCharacters:buffer];
    buffer[length] = '\0';
    trieNodeClass = ((OFTrie *)head)->isa;
    if (head->count) {
	while ((to = trieFindChild(attachTo, *ptr))) {
            if (((OFTrie *)to)->isa != trieNodeClass) {
		OFTrieBucket *existing = (OFTrieBucket *)to;
		OFTrieNode *end = attachTo;
		unichar *existingPtr = existing->lowerCharacters - 1;
		unichar *ptrPosition = ptr;

                [existing retain];
		if (!*ptrPosition) {
		    ptr = ptrPosition;
		    break;
		}
		do {
                    OFTrieNode *new;

                    new = [[OFTrieNode allocWithZone:[end zone]] init];
		    [end addChild:new withCharacter:*ptr];
		    [end addChild:new withCharacter:upperBuffer[(ptr-buffer)]];
		    [new release];
		    end = new;
		    ptr++;
		    existingPtr++;
		} while (*ptr && *ptr == *existingPtr);

		if (*existingPtr || *ptr) {
                    unsigned int offset;

                    offset = existingPtr - existing->lowerCharacters;
		    if (*existingPtr) {
                        unsigned int existingLength;

			[end addChild:existing withCharacter:*existingPtr];
			[end addChild:existing 
			    withCharacter:existing->upperCharacters[offset]];
		
                        existingLength = 0;
                        while (*++existingPtr != '\0')
                            existingLength++;
			offset++;
			[existing setRemainingLower:existing->lowerCharacters + offset upper:existing->upperCharacters + offset length:existingLength];
		    } else {
			[end addChild:existing withCharacter:0];
			offset++;
			[existing setRemainingLower:existing->lowerCharacters + offset upper:existing->upperCharacters + offset length:0];
		    }
		    attachTo = end;
		} else {
		    ptr = ptrPosition;
		}
                [existing release];
		break;
	    }
	    attachTo = to;
	    ptr++;	
	}
    }
    [attachTo addChild:bucket withCharacter:upperBuffer[ptr - buffer]];
    [attachTo addChild:bucket withCharacter:*ptr++];
    if (caseSensitive)
	[bucket setRemainingLower:ptr upper:ptr length:length - (ptr - buffer)];
    else
	[bucket setRemainingLower:ptr upper:upperBuffer + (ptr - buffer) length:length - (ptr - buffer)];

    if (useMalloc) {
	NSZoneFree(NULL, buffer);
	NSZoneFree(NULL, upperBuffer);
    }
}

- (OFTrieBucket *)bucketForString:(NSString *)aString;
{
    unsigned int length;
    unichar *buffer, *ptr;
    OFTrieNode *currentNode;
    Class trieNodeClass;
    BOOL useMalloc;

    if (head->count == 0)
	return nil;

    length = [aString length];
    useMalloc = (length + 1) * sizeof(*buffer) >= SAFE_ALLOCA_SIZE;
    if (useMalloc)
	buffer = (unichar *)NSZoneMalloc(NULL, (length + 1) * sizeof(*buffer));
    else
	buffer = (unichar *)alloca((length + 1) * sizeof(*buffer));
    [aString getCharacters:buffer];
    buffer[length] = 0;
    ptr = buffer;
    currentNode = head;
    trieNodeClass = ((OFTrie *)head)->isa;
    while ((currentNode = trieFindChild(currentNode, *ptr++))) {
	if (((OFTrie *)currentNode)->isa != trieNodeClass) {
	    OFTrieBucket *test;
	    unichar *lowerPtr, *upperPtr;

	    test = (OFTrieBucket *)currentNode;
	    lowerPtr = test->lowerCharacters;
	    upperPtr = test->upperCharacters;
	    if (!ptr[-1] && !*lowerPtr) {
		if (useMalloc)
		    NSZoneFree(NULL, buffer);
		return test;
	    }
	    while (*ptr) {
		if (*ptr != *lowerPtr && *ptr != *upperPtr)
		    return nil;
		lowerPtr++, upperPtr++, ptr++;
	    }
	    if (useMalloc)
		NSZoneFree(NULL, buffer);
	    return *lowerPtr ? nil : test;
	}
    }
    if (useMalloc)
	NSZoneFree(NULL, buffer);
    return nil;
}

- (OFTrieNode *)headNode;
{
    return head;
}

// Debugging

- (NSMutableDictionary *)debugDictionary;
{
    NSMutableDictionary *debugDictionary;

    debugDictionary = [super debugDictionary];
    [debugDictionary setObject:head forKey:@"head"];
    return debugDictionary;
}

@end
